/******************************************************************************
* Module:
*	LinkedList Reversal
*
* Description:
*	This implements a simple linked list, and a reversal function for it.
*
*	A startup thinks that real programmers are rare, so I slapped this
*	together as a demonstration.  It really isn't rocket science, and
*	I find it highly doubtful that most programmers couldn't do this
*	if given the peace and quiet to do so.
*
*	Elasped time: About 40 minutes.
*
* Author:
*	John W. M. Stevens
******************************************************************************/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>

#include	"LinkedList.h"

/*-----------------------------------------------------------------------------
| Function:
|	LinkedList_Reverse
|
| Description:
|	Reverses the sinly linked list specified in LinkedList.h.
|
|	Approximate performance is O( n ).
|
| Parameters:
|	root - The root node of the list.
|
| Returns:
|	Returns the reversed list.
-----------------------------------------------------------------------------*/

static
ListElement_t	*
LinkedList_Reverse(
	ListElement_t	*root )
{
	auto	ListElement_t	*current;
	auto	ListElement_t	*previous;
	auto	ListElement_t	*next;

	/*	No list?  Return no list.	*/
	if (root == 0)
		return 0;

	/*	Initialize linked list traversal.	*/
	current  = root;
	previous = 0;

	/*	Run list, reversing as we go.	*/
	do
	{
		/*	Get pointer to next list element.	*/
		next = current->next;

		/*	Repoint current at previous.	*/
		current->next = previous;

		/*	Shift by one.	*/
		previous = current;
		current  = next;
	} while (current != 0);

	/*	Return new head of list.	*/
	return previous;
}

/*-----------------------------------------------------------------------------
| Function:
|	Test_BuildList
|
| Description:
|	Build a short, simple little linked list to reverse for testing
|	purposes.  If you haven't tested it, you aren't done.
|
|	Obviously, there are a number of tests that should be written
|	besides this one, including the various boundary tests (empty
|	list, one element list, for example).
|
| Parameters:
|	None.
|
| Returns:
|	Returns the new list.
-----------------------------------------------------------------------------*/

static
ListElement_t	*
Test_BuildList(
	void )
{
	register	int				i;
	auto		int				nodeSize;
	auto		ListElement_t	*root;
	auto		ListElement_t	*node;
	auto		ListElement_t	*previous;

	/*	Test data for a four element list.	*/
	static	char	*testList[] = {
		"First line.",
		"Second line.",
		"Third line.",
		"Fourth line.",
		0
	};

	/*	For the number of test strings, do:	*/
	for (i        = 0,
		 previous = 0;
		 
		 testList[ i ];
		 
		 i++)
	{
		/*	Calculate node size.	*/
		nodeSize = sizeof( ListElement_t ) +
			strlen(
				testList[ i ] ) + 1;

		/*	Can we allocate a node?	*/
		if ( (node =
			calloc(
				1,
				nodeSize )) == 0)
		{
			/*	No.  Out of memory.  Report error and exit.	*/
			fprintf(
				stderr,
				"%s %d : ERROR - Out of memory.\n",
				__FILE__,
				__LINE__ );
			exit(
				1 );
		}

		/*	Initialize node.	*/
		node->sizeInBytes = nodeSize;
		strcpy(
			node->data,
			testList[ i ] );

		/*	Save root.	*/
		if (i == 0)
			root = node;
		else
			previous->next = node;

		/*	Save previous.	*/
		previous = node;
	}

	/*	Return root.	*/
	return root;
}

/*-----------------------------------------------------------------------------
| Function:
|	Debug_PrintList
|
| Description:
|	Grossly simple list printer.  This assumes that the data block contains
|	a C string, so don't get careless about using this!
|
| Parameters:
|	root - The root node of the singly linked list.
|
| Returns:
|	Nothing.
-----------------------------------------------------------------------------*/

void
Debug_PrintList(
	ListElement_t	*root )
{
	for ( ; root; root = root->next )
		printf(
			"%s\n",
			root->data );
}

/*=============================================================================
||	Drive it.
=============================================================================*/

int
main(
	void )
{
	auto	ListElement_t	*root;
	auto	ListElement_t	*reversed;

	/*	Run test function.	*/
	root =
		Test_BuildList();

	/*	Show list before reversal.	*/
	printf(
		"Linked List:\n"
		"------------\n" );
	Debug_PrintList(
		root );

	/*	Reverse list.	*/
	reversed =
		LinkedList_Reverse(
			root );

	/*	Show list after reversal.	*/
	printf(
		"\n"
		"Reversed List:\n"
		"--------------\n" );
	Debug_PrintList(
		reversed );

	/*	Return OK.	*/
	exit(
		0 );
}
